import java.io.*;
import java.util.*;

public class PDB implements PDBParser {
	static ArrayList<String> Reijmers_atoms=new ArrayList<String>(Arrays.asList("C3'", "C4'", "C5'", "O3'", "O5'", "P"));	//heavy atoms used in Reijmers' paper

	public HashMap<String, ArrayList<Residue>> chain_res;	//key: chain; value: residue
	public HashMap<String, ArrayList<Atom>> chain_atom;	//key: chain; value: coordinates of residue;
	
	public HashMap<String, ArrayList<Residue>> get_chain_res() {
		return chain_res;
	}

	public HashMap<String, ArrayList<Atom>> get_chain_atom() {
		return chain_atom;
	}
	
	public PDB(File fname, String _chainID) throws IOException{
		
		chain_res=new HashMap<String, ArrayList<Residue>>();
		chain_atom=new HashMap<String, ArrayList<Atom>>();
		HashSet<String> chain_TER=new HashSet<String>();	//the terminated chains
		
		String line;
		
		String chainID = null;
		char icode='\0';
		int seqnum=0;
		int sn=0;
		String symbol;
		double x=0., y=0., z=0.;
		String atom;
		ResID rid;
		ResID cur_rid=new ResID(null, -1, '\0');	//set current resid to null

		BufferedReader in=new BufferedReader(new FileReader(fname));
		
		while((line=in.readLine())!=null){
			if(line.startsWith("ENDMDL")) break;	//only consider the first model
			if(line.startsWith("TER")) { chain_TER.add(chainID); cur_rid=new ResID(null, -1, '\0');}	//add the terminated chain to chain_TER and set the current residue to null
			if(line.startsWith("ATOM") || line.startsWith("HETATM")){
				chainID = "" + line.charAt(21);
				if(chain_TER.contains(chainID)==true) continue;		//if chain is terminated, do nothing
				if(chain_res.get(chainID)==null) chain_res.put(chainID, new ArrayList<Residue>());	//create the residue list and coordinates list for the chain (first appearance)
				if(chain_atom.get(chainID)==null) chain_atom.put(chainID, new ArrayList<Atom>());	
				
				symbol=line.substring(17, 20).trim();
				seqnum=Integer.valueOf(line.substring(22, 26).trim());
				sn=Integer.valueOf(line.substring(6, 11).trim());
				icode=line.charAt(26);
				atom=line.substring(12, 16).trim();
				x=Double.valueOf(line.substring(30, 38).trim());
				y=Double.valueOf(line.substring(38, 46).trim());
				z=Double.valueOf(line.substring(46, 54).trim());
				
				rid=new ResID(chainID, seqnum, icode);
				
				if(rid.equals(cur_rid) == false) {	//new residue
					chain_res.get(chainID).add(new Residue(rid, symbol));
					cur_rid=rid;
				}
				chain_atom.get(chainID).add(new Atom(sn, new Residue(cur_rid, symbol), atom, new Point(x, y, z)));	//add the new atom
			}
		}
		in.close();
	}
	
	//return the RNA sequence for chain (N for unknown residue)
	public String get_chain_seq(String chainID){
		String seq="";
		
		for(Residue R: chain_res.get(chainID)){
			if(R.symbol.equals("A") || R.symbol.equals("C")|| R.symbol.equals("G") || R.symbol.equals("U"))
				seq+=R.symbol;
			else
				seq+="N";
		}
		return seq;
	}

	/*
	public ArrayList<Point> get_chain_coord(Character chainID){
		ArrayList<Point> coord =new ArrayList<Point>();
		for(ArrayList<Point> L: chain_coord.get(chainID)){
			Point P=Geom.centroid(L);
			coord.add(P);
		}
		return coord;
	}
	*/
	
	public ArrayList<Point> get_chain_centroid(String chainID){
		ArrayList<ArrayList<Point>> coord = new ArrayList<ArrayList<Point>>();
		ArrayList<Point> centroid = new ArrayList<Point>();
		
		ResID cur_rid=new ResID(null, -1, '\0');
		
		for(Atom A: chain_atom.get(chainID)){
			if (A.res.rid.equals(cur_rid)==false) {coord.add(new ArrayList<Point>());  cur_rid=A.res.rid;};
			if (Reijmers_atoms.contains(A.atom)) coord.get(coord.size()-1).add(A.coord); 
		}
		
		for(ArrayList<Point> L: coord){
			centroid.add(Geom.centroid(L));
		}
		
		return centroid;
	}
	
	public static void main(String[] args) throws Exception{
		PDB pdb=new PDB(new File("/home/xiaoli/software/ori_star/STAR3D_source/PDB/4bw0.pdb"),"A");
		for(Residue A: pdb.chain_res.get("A")) System.out.println(A);
	}
}
